# FrequencySweep.jl

The FrequencySweep library is tailored towards the calculation of frequency response functions for systems of ordinary differential equations that are difficult or impractical to compute by typical analytical or numerical means (e.g. nonlinear models such as Duffing's equation).  

As an illustration, consider Duffing's equation.

$\ddot{x} + \delta \dot{x} + \alpha x + \beta x^{3} = \gamma \sin(\omega t)$

To use this library the equation needs to be written in terms of two first order equations.
```julia
function duffing_eqn(du, u, p, f, t)
    alpha, beta, gamma, delta = p
    du[1] = u[2]
    du[2] = gamma * sin(f * t) - delta * u[2] - alpha * u[1] - beta * u[1]^3
end
```

The arguments to the function are as follows:
- *du:* This is an output array where the output of each first order equation should be written.
- *u:* This is an input array the same size as *du* providing the current estimates for each state variable.
- *p:* A structure of user-defined type that can be used to pass model parameters.
- *f:* The oscillation frequency the solver is analyzing.  This value will be supplied in the same units as supplied by the user.
- *t:* The independent variable, typically time.

Duffing's equation exhibits the jump phenomenon near resonance.  The nature of the jump behavior is controlled by the parameters utilized.  Setting $\alpha = \gamma = 1$, $\delta = 0.1$, and $\beta = 0.5$ produces the following frequency response.

![](Duffing_Analytical.png?raw=true)

The frf_sweep function computes an estimate to the frequency response function by solving the system of equations supplied by the user for a predetermined number of oscillations.  The solver then computes the amplitude and phase of the last few cycles at end of the solution.  These constraints can be defined by the user to better control the process.  The solver computes the solution at each supplied frequency point sweeping in the direction defined by the user.  The solver uses the final solution point of the previous solution as initial conditions for the next frequency point.

Here is how it all looks in Julia.
```julia
f1 = 0.5:0.05:2.0           # sweeping forward in frequency
f2 = 2.0:-0.05:0.5          # sweeping backward in frequency

sweep1 = frf_sweep(
    duffing_eqn,            # function containing the equations to solve
    [0.0, 0.0],             # initial condition vector
    f1                      # frequency points
)
sweep2 = frf_sweep(
    duffing_eqn,
    [0.0, 0.0],
    f2
)
```
The solution returned from frf_sweep is an M-by-N matrix where M is the number of frequency points and N is the number of equations defined.  The solution is complex-valued such that each column supplies both magnitude and phase information for each equation.  For the Duffing model utilized above, the frequency response of the first equation is as follows.

![](Duffing.png?raw=true)

A word of caution however.  Ensure that only non-zero frequency values are specified as the frequency value determines the length of solution as the solver uses the frequency values to determine the length of the ODE solution.

## Function Documentation
```@docs
frf_sweep
```
