# FrequencySweepCore.jl

using DifferentialEquations

include("HelperRoutines.jl")
include("Spectral.jl")

"""
rst = frf_sweep(fcn, init, freq)

## Description

The frf_sweep function computes the frequency response functions for a system of
ordinary differential equations (ODE's) by means of "sweeping" across a range of 
excitation frequencies.  The algorithm solves the system of ODE's at each 
excitation frequency in the supplied range.  It uses the solution of these 
equations to determine the amplitude and phase of the resulting wavefuorm once
it has sufficiently damped out the transient component of the solution.  The 
amplitude and phase are determined by a discrete Fourier transform of the last
N cycles of the solution waveform for each ODE.

*fcn:*
The solver requires the user to define a function supplying the ODE's to 
integrate.  This function must be defined as follows.

```julia
function fcn(du, u, p, omega, t)
    # User code goes here
end
```

The arguments to the function are as follows:
- *du:* This is an output array where the output of each first order equation should be written.
- *u:* This is an input array the same size as *du* providing the current estimates for each state variable.
- *p:* A structure of user-defined type that can be used to pass model parameters.
- *f:* The oscillation frequency the solver is analyzing.  This value will be supplied in the same units as supplied by the user.
- *t:* The independent variable, typically time.

*init:*
The user must also supply an initial set of initial conditions for the solution 
of the ODE's.  This set of initial conditions is only used for the solution to 
the first frequency value in the range as the solver will update the initial 
condition vector for subsequent frequency value solutions.

*freq:*
The user must also supply a range of frequency values at which to compute the 
frequency responses.  Ensure that there are no zero-valued frequencies in the 
range as the solver does not accept such a value.  The solver utilizes the 
frequency value as a method of determining the length of the ODE solution; 
therefore, smaller frequency values result in longer solutions.  A note about 
units: It is recommended that the frequency values be supplied with units of 
radians per unit time; however, whatever units are provided will be the same 
units used for the value passed to *fcn*.

The function returns a named tuple with two fields:
- frequency: An M-element array containing the frequency points at which the 
solution was computed.
- response: An M-by-N matrix where M is the number of frequency points 
and N is the number of ODE's being solved.  The result is complex-valued such 
that each column of the matrix describes both the amplitude and phase of each 
equation.

The user may supply a set of optional parameters to control both the ODE solver 
and the calculation of amplitude and phase for each equation.  The optional 
parameters are as follows.

### ODE Solver Optional Arguments
- alg_hints: Any hints to pass to the ODE solver (e.g. :stiff).
- reltol: The relative solution tolerance.
- abstol: The absolute solution tolerance.
- dtmax: The maximum allowable step size in the independent variable.
- maxiters: The maximum number of iterations to allow.
- modelparams: A container for arguments to pass to *fcn*.
- solver: The ODE solver to utilize.  The default is Tsit5().  See 
DifferentialEquations.jl for all available solvers.

### Additional Optional Arguments
- pointspercycle: Defines the number of solution points for each cycle of the solution.
- cyclecount: Defines the number of solutions to use in the calculation of the amplitude and phase.
- transientcount: Defines how many cycles to let pass for the transient to sufficiently decay.
- windowsize: Defines the size of window to utilize for the discrete Fourier transform.
- windowfunction: Defines the window function to utilize for the discrete Fourier transform.  The function should take the same form as the windowing functions from the DSP.jl library.
- zerotol: Provides a tolerance of closeness to zero for frequency values. Frequency values whose magnitude is below this threshold are ignored during the analysis.

## Example
The following example illustrates how to use the function to obtain both an 
ascending and descending solution to Duffing's equation.
```julia
# Duffing's Equation
function duffing_eqn(du, u, p, omega, t)
    du[1] = u[2]
    du[2] = gamma * sin(omega * t) - delta * u[2] - alpha * u[1] - beta * u[1]^3
end

# Frequency Points
f1 = 0.5:0.05:2.0
f2 = 2.0:-0.05:0.5

# Perform the sweep
sweep1 = frf_sweep(
    duffing_eqn,
    [0.0, 0.0],
    f1
)
sweep2 = frf_sweep(
    duffing_eqn,
    [0.0, 0.0],
    f2
)
```
"""
function frf_sweep(
    fcn::Function,
    init::AbstractVector{T},
    freq::AbstractVector{T};
    args...
) where T <: Real
    # Define the function to solve
    function diffeqfun!(du_, u_, p_, t_)
        # Extract the necessary parameters
        omega_, params_ = p_

        # Evaluate the user-supplied function
        fcn(du_, u_, params_, omega_, t_)
    end

    # Extract user options
    opts = extract_options(args)
    nt = opts.transientcount
    nc = opts.cyclecount
    ppc = opts.pointspercycle
    zt = opts.zerotol

    # Set up an output matrix
    ndof = length(init)
    nfreq = length(freq)
    rst = zeros(ComplexF64, nfreq, ndof)
    f = zeros(T, nfreq)

    # Allocate space for a time vector
    tvec = zeros(T, ppc * nc)

    # Cycle over each frequency
    j = 0
    for omega in freq
        # Check for a zero frequency.  If found, warn the user and cycle to the 
        # next frequency
        if abs(omega) < zt
            warn("$omega was found to be too close to zero.  As a result, this frequency is being skipped.  Adjust zerotol if this frequency must be included in the results.")
            continue
        end

        # Define the solution time vector - only need the points of concern
        # for the steady-state portion of the solution.
        fill_time_vector!(nt, ppc, omega, tvec)
        
        # Establish parameters to pass to the solver
        params = (omega, opts.modelparams)

        # Set-up the solver
        prob = ODEProblem(diffeqfun!, init, [0.0, tvec[end]], params)

        # Solve the ODE's
        sol = solve(
            prob,
            opts.solver,
            saveat = tvec,
            reltol = opts.reltol,
            abstol = opts.abstol,
            dtmax = opts.dtmax,
            maxiters = opts.maxiters,
            alg_hints = opts.alg_hints
        )

        # Check for solution errors or other issues that may cause problems from
        # here on out
        if sol.retcode != :Default && sol.retcode != :Success
            # The solver did not succeed at this frequency.  Warn the user and 
            # move on to the next frequency
            warn("The ODE solver did not succeed in computing a full solution at frequency $omega.  This frequency is being skipped.")
            continue
        end

        # Index the counter variable
        j = j + 1

        # Cycle over each solution and compute the amplitude and phase
        for i in 1:ndof
            x = spectrum(
                opts.windowfunction,
                opts.windowsize,
                sol[i,:]
            )
            ind = argmax(abs.(x[2:end])) + 1 # Guard against a DC value being chosen
            rst[j,i] = x[ind]
        end
        
        # Store the frequency value
        f[j] = omega

        # Reset the initial conditions for the next go around
        init = sol[:,end]
    end

    # End
    return (
        frequency = f[1:j],
        response = rst[1:j,:]
    )
end

# ------------------------------------------------------------------------------
# Internal Function:
#
# This function populates an N-element time vector.  This is an in-place 
# function that does not require additional array allocation.  Notice, this 
# array will only include the steady state cycles though as only the steady
# state time points need to be saved.
function fill_time_vector!(
    nTransientCycles::Int,  # number of transient cycles
    pointsPerCycle::Int,    # number of time points per each cycle
    omega::T,               # frequency
    t::AbstractVector{T}    # output where the new time array is written
) where T <: Real
    npts = length(t)
    period = 2 * pi / abs(omega)
    dt = period / (pointsPerCycle - 1)
    t[1] = nTransientCycles * period + dt
    for i in 2:npts
        t[i] = t[i-1] + dt
    end
end
