var documenterSearchIndex = {"docs":
[{"location":"#FrequencySweep.jl","page":"FrequencySweep.jl","title":"FrequencySweep.jl","text":"","category":"section"},{"location":"","page":"FrequencySweep.jl","title":"FrequencySweep.jl","text":"The FrequencySweep library is tailored towards the calculation of frequency response functions for systems of ordinary differential equations that are difficult or impractical to compute by typical analytical or numerical means (e.g. nonlinear models such as Duffing's equation).  ","category":"page"},{"location":"","page":"FrequencySweep.jl","title":"FrequencySweep.jl","text":"As an illustration, consider Duffing's equation.","category":"page"},{"location":"","page":"FrequencySweep.jl","title":"FrequencySweep.jl","text":"ddotx + delta dotx + alpha x + beta x^3 = gamma sin(omega t)","category":"page"},{"location":"","page":"FrequencySweep.jl","title":"FrequencySweep.jl","text":"To use this library the equation needs to be written in terms of two first order equations.","category":"page"},{"location":"","page":"FrequencySweep.jl","title":"FrequencySweep.jl","text":"function duffing_eqn(du, u, p, f, t)\r\n    alpha, beta, gamma, delta = p\r\n    du[1] = u[2]\r\n    du[2] = gamma * sin(f * t) - delta * u[2] - alpha * u[1] - beta * u[1]^3\r\nend","category":"page"},{"location":"","page":"FrequencySweep.jl","title":"FrequencySweep.jl","text":"The arguments to the function are as follows:","category":"page"},{"location":"","page":"FrequencySweep.jl","title":"FrequencySweep.jl","text":"du: This is an output array where the output of each first order equation should be written.\nu: This is an input array the same size as du providing the current estimates for each state variable.\np: A structure of user-defined type that can be used to pass model parameters.\nf: The oscillation frequency the solver is analyzing.  This value will be supplied in the same units as supplied by the user.\nt: The independent variable, typically time.","category":"page"},{"location":"","page":"FrequencySweep.jl","title":"FrequencySweep.jl","text":"Duffing's equation exhibits the jump phenomenon near resonance.  The nature of the jump behavior is controlled by the parameters utilized.  Setting alpha = gamma = 1, delta = 01, and beta = 05 produces the following frequency response.","category":"page"},{"location":"","page":"FrequencySweep.jl","title":"FrequencySweep.jl","text":"(Image: )","category":"page"},{"location":"","page":"FrequencySweep.jl","title":"FrequencySweep.jl","text":"The frf_sweep function computes an estimate to the frequency response function by solving the system of equations supplied by the user for a predetermined number of oscillations.  The solver then computes the amplitude and phase of the last few cycles at end of the solution.  These constraints can be defined by the user to better control the process.  The solver computes the solution at each supplied frequency point sweeping in the direction defined by the user.  The solver uses the final solution point of the previous solution as initial conditions for the next frequency point.","category":"page"},{"location":"","page":"FrequencySweep.jl","title":"FrequencySweep.jl","text":"Here is how it all looks in Julia.","category":"page"},{"location":"","page":"FrequencySweep.jl","title":"FrequencySweep.jl","text":"f1 = 0.5:0.05:2.0           # sweeping forward in frequency\r\nf2 = 2.0:-0.05:0.5          # sweeping backward in frequency\r\n\r\nsweep1 = frf_sweep(\r\n    duffing_eqn,            # function containing the equations to solve\r\n    [0.0, 0.0],             # initial condition vector\r\n    f1                      # frequency points\r\n)\r\nsweep2 = frf_sweep(\r\n    duffing_eqn,\r\n    [0.0, 0.0],\r\n    f2\r\n)","category":"page"},{"location":"","page":"FrequencySweep.jl","title":"FrequencySweep.jl","text":"The solution returned from frf_sweep is an M-by-N matrix where M is the number of frequency points and N is the number of equations defined.  The solution is complex-valued such that each column supplies both magnitude and phase information for each equation.  For the Duffing model utilized above, the frequency response of the first equation is as follows.","category":"page"},{"location":"","page":"FrequencySweep.jl","title":"FrequencySweep.jl","text":"(Image: )","category":"page"},{"location":"","page":"FrequencySweep.jl","title":"FrequencySweep.jl","text":"A word of caution however.  Ensure that only non-zero frequency values are specified as the frequency value determines the length of solution as the solver uses the frequency values to determine the length of the ODE solution.","category":"page"},{"location":"#Function-Documentation","page":"FrequencySweep.jl","title":"Function Documentation","text":"","category":"section"},{"location":"","page":"FrequencySweep.jl","title":"FrequencySweep.jl","text":"frf_sweep","category":"page"},{"location":"#FrequencySweep.frf_sweep","page":"FrequencySweep.jl","title":"FrequencySweep.frf_sweep","text":"rst = frf_sweep(fcn, init, freq)\n\nDescription\n\nThe frf_sweep function computes the frequency response functions for a system of ordinary differential equations (ODE's) by means of \"sweeping\" across a range of  excitation frequencies.  The algorithm solves the system of ODE's at each  excitation frequency in the supplied range.  It uses the solution of these  equations to determine the amplitude and phase of the resulting wavefuorm once it has sufficiently damped out the transient component of the solution.  The  amplitude and phase are determined by a discrete Fourier transform of the last N cycles of the solution waveform for each ODE.\n\nfcn: The solver requires the user to define a function supplying the ODE's to  integrate.  This function must be defined as follows.\n\nfunction fcn(du, u, p, omega, t)\n    # User code goes here\nend\n\nThe arguments to the function are as follows:\n\ndu: This is an output array where the output of each first order \n\nequation should be written.\n\nu: This is an input array the same size as du providing the current \n\nestimates for each state variable.\n\np: A structure of user-defined type that can be used to pass model \n\nparameters.\n\nf: The oscillation frequency the solver is analyzing.  This value will \n\nbe supplied in the same units as supplied by the user.\n\nt: The independent variable, typically time.\n\ninit: The user must also supply an initial set of initial conditions for the solution  of the ODE's.  This set of initial conditions is only used for the solution to  the first frequency value in the range as the solver will update the initial  condition vector for subsequent frequency value solutions.\n\nfreq: The user must also supply a range of frequency values at which to compute the  frequency responses.  Ensure that there are no zero-valued frequencies in the  range as the solver does not accept such a value.  The solver utilizes the  frequency value as a method of determining the length of the ODE solution;  therefore, smaller frequency values result in longer solutions.  A note about  units: It is recommended that the frequency values be supplied with units of  radians per unit time; however, whatever units are provided will be the same  units used for the value passed to fcn.\n\nThe function outputs an M-by-N matrix where M is the number of frequency points  and N is the number of ODE's being solved.  The result is complex-valued such  that each column of the matrix describes both the amplitude and phase of each  equation.\n\nThe user may supply a set of optional parameters to control both the ODE solver  and the calculation of amplitude and phase for each equation.  The optional  parameters are as follows.\n\nODE Solver Optional Arguments\n\nalg_hints: * Any hints to pass to the ODE solver (e.g. :stiff).\nreltol: * THe relative solution tolerance.\nabstol: * The absolute solution tolerance.\ndtmax: * The maximum allowable step size in the independent variable.\nmaxiters: * The maximum number of iterations to allow.\nmodelparams: * A container for arguments to pass to fcn.\n\nAdditional Optional Arguments\n\npointspercycle: * Defines the number of solution points for each cycle \n\nof the solution.\n\ncyclecount: * Defines the number of solutions to use in the calculation \n\nof the amplitude and phase.\n\ntransientcount: * Defines how many cycles to let pass for the transient \n\nto sufficiently decay.\n\nwindowsize: * Defines the size of window to utilize for the discrete \n\nFourier transform.\n\nwindowfunction: * Defines the window function to utilize for the \n\ndiscrete Fourier transform.  The function should take the same form as the  windowing functions from the DSP.jl library.\n\nExample\n\nThe following example illustrates how to use the function to obtain both an  ascending and descending solution to Duffing's equation.\n\n# Duffing's Equation\nfunction duffing_eqn(du, u, p, omega, t)\n    du[1] = u[2]\n    du[2] = gamma * sin(omega * t) - delta * u[2] - alpha * u[1] - beta * u[1]^3\nend\n\n# Frequency Points\nf1 = Array(0.5:0.05:2.0)\nf2 = Array(2.0:-0.05:0.5)\n\n# Perform the sweep\nsweep1 = frf_sweep(\n    duffing_eqn,\n    [0.0, 0.0],\n    f1\n)\nsweep2 = frf_sweep(\n    duffing_eqn,\n    [0.0, 0.0],\n    f2\n)\n\n\n\n\n\n","category":"function"}]
}
